part of TeamSpeak;

/**
 * Represents a user that is or was active on a server that is or was connected.
 * 
 * Please note that all fields are updated only once when the client is first
 * seen and then only after update is called.
 */

class User extends TeamSpeakEntity {
  /// Unique ID used by the TeamSpeak-Server's database to identify a client
  final int database_id;
  
  /// Client-ID. Client-IDs may change when a client disconnects and connects
  /// again. Client-IDs are only used to identify clients that are currently online
  /// since every id can only be used by one client at a time.
  final int clid;
  
  /// Checks whether the client is currently online. 
  /// Note that the client may be online on another (virtual) server when this
  /// returns false.
  bool is_online = true;
  
  /// ID of the channel the client is in (or was).
  int get cid => _cid; int _cid;
  
  /// Unique identifier string used by the TeamSpeak-Server's database to identify a client
  String get unique_identifier => _unique_identifier; String _unique_identifier; 
  
  /// Time the user has spent idling on the server (not communicating).
  int get idle_time => _idle_time; int _idle_time;
  
  /// Nickname the user uses
  String get nickname => _nickname; String _nickname;
  
  /// Version of the client the user uses to connect to the server
  String get version => _version; String _version;
  
  /// Platform the user's client is running on (for example Windows).
  String get platform => _platform; String _platform;
  
  /// Returns true if the user has muted his microphone or if it's deactivated.
  bool get input_muted => _input_muted; bool _input_muted;
  
  /// Returns true if the user has muted his speakers / phones or if they are deactivated.
  bool get output_muted => _output_muted; bool _output_muted;
  
  /// Channel the client joins automatically after connecting to the server
  /// (usually the server's default channel)
  Channel get default_channel => _default_channel; Channel _default_channel;
  
  /// Returns true if the user is currently recording audio
  bool get is_recording => _is_recording; bool _is_recording;
  
  /// Returns the time the user was created. This is usually the time
  /// the user connected to the server for the first time.
  DateTime get created => _created; DateTime _created;

  /// Returns the time the user established his most recent connection to the server
  DateTime get lastconnected => _lastconnected; DateTime _lastconnected;
  
  /// Returns the amount of times the user connected to this server
  int get totalconnections => _totalconnections; int _totalconnections;
  
  /// Returns whether the user is currently marked as being away.
  bool get away => _away; bool _away;
  
  /// If [away] is set, this holds the user's individual away message / reason 
  /// (if defined).
  String away_message; String _away_message;
  
  /// IP address or hostname the client is connecting from.
  String ip; String _ip;
  
  /// Returns true if the client is connected using the ServerQuery-protocol
  /// (like us).
  bool get is_serverquery_client => _type > 0; int _type;
  
  /// Total amount of packets sent *by the server* during the client's current connection.
  int get packets_sent_total => _packets_sent_total; int _packets_sent_total;
  
  /// Total amounf of packets received *by the server* during the client's current connection.
  int get packets_received_total => _packets_received_total; int _packets_received_total;
  
  /// Bytes sent to the client in the last second before [update] was called.
  int get bandwidth_sent_last_second_total => _bandwidth_sent_last_second_total; int _bandwidth_sent_last_second_total;
  
  /// Bytes received from the client in the last second before [update] was called.
  int get bandwidth_received_last_second_total => _bandwidth_received_last_second_total; int _bandwidth_received_last_second_total;
  
  /// Returns the channel the client is in (or was)
  Channel get channel => _channel; Channel _channel;
  
  /// Updates the information about this client. Is done automatically when
  /// the client is discovered for the first time, but can be done manually
  /// afterwards to gain new data.
  Future<User> update(){
    return _client.send(new Command("clientinfo", parameters: {"clid": clid}))
        .then((Answer answer){
          final Map<String, String> info = answer.items[0];
          _unique_identifier = info["client_unique_identifier"];
          _idle_time = int.parse(info["client_idle_time"]);
          _cid = int.parse(info["cid"]);
          _nickname = info["client_nickname"];
          if(!server.channels.containsKey(_cid)){
             throw new StateError("${_nickname} is in a channel that does not exist on the server-mirror.");
          }
          _version = info["client_version"];
          _platform = info["client_platform"];
          _input_muted = info["client_input_muted"] == "1";
          _output_muted = info["client_output_muted"] == "1";
          _is_recording = info["client_is_recording"] == "1";
          _created = new DateTime.fromMillisecondsSinceEpoch(int.parse(info["client_created"])*1000);
          _lastconnected = new DateTime.fromMillisecondsSinceEpoch(int.parse(info["client_lastconnected"])*1000);
          _totalconnections = int.parse(info["client_totalconnections"]);
          _away = info["client_away"] == "1";
          _away_message = info["client_away_message"];
          _ip = info["connection_client_ip"];
          _type = int.parse(info["client_type"]);
          _packets_sent_total = int.parse(info["connection_packets_sent_total"]);
          _packets_received_total = int.parse(info["connection_packets_received_total"]);
          _bandwidth_received_last_second_total = int.parse(info["connection_bandwidth_received_last_second_total"]);
          _bandwidth_sent_last_second_total = int.parse(info["connection_bandwidth_sent_last_second_total"]);         
        
          try {
            final int default_channel_id = int.parse(info["client_default_channel"].substring(1));
            _default_channel = (server.channels.containsKey(default_channel_id) ? server.channels[default_channel_id] : null);
          } catch(e){
            _default_channel = null;
          }
          
          _channel = server.channels[_cid];
          
          return this;
        }, onError: (Error e){
          is_online = false;
        });
  }
  
  User(Server server, _Client _client, int this.clid, int this.database_id) 
    : super(server, _client);
}