/*
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 42):
 * <timklge@gmail.com> wrote this file. As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return Tim Kluge
 * ----------------------------------------------------------------------------
 */

part of TeamSpeak;

/**
 * Represents a TeamSpeak-ServerQuery "packet" that can be send over network.
 * A packet may be a [Command] sent by the client (for example clientlist) or
 * an [Answer] which is sent by the server to reply to a Command.
 * 
 * This class provides low-level methods to work with the protocol.
 */

abstract class Packet {
  /// Transition table used to escape strings
  static Map<String, String> transTable = {
   r"/": r"\/",
   r" ": r"\s",
   r"|": r"\p",
  
   "\x07": r"\a",
  
   "\b": r"\b",
   "\f": r"\f",
   "\n": r"\n",
   "\r": r"\r",
   "\t": r"\t",
   "\v": r"\v",
  
   r"\": r"\\"
  };
  
  /// Parses a string that holds key-value-pairs. These are escaped according to the
  /// TeamSpeak-Protocol.
  static Map<String, String> parseItem(String pItem){
    Map<String, String> lMap = new Map();
    pItem.split(" ").forEach((String pString){
      final int lIndex = pString.indexOf("=");
      if(lIndex < 0){ // No = in this - no key-value-pair, key with empty value
        lMap[pString] = "";
      } else {
        String lKey = unescape(pString.substring(0, lIndex));
        String lValue = unescape(pString.substring(pString.indexOf("=")+1));
        lMap[lKey] = lValue;
      }
    });
    return lMap;
  }
  
  /// Escapes a string so it can be transmitted according to the TeamSpeak-protocol.
  static String escape(String pString){
    String lString = pString;
    transTable.forEach((String pKey, String pValue){
      lString = lString.replaceAll(pKey, pValue);
    });
    return lString;
  }
  
  /// Unescapes a string that was previously escaped (using [escape]).
  static String unescape(String pString){
    String lString = pString;
    transTable.forEach((String pKey, String pValue){
      lString = lString.replaceAll(pValue, pKey);
    });
    return lString;
  }
}

class Command extends Packet {
  final String command;

  Map<String, dynamic> _parameters = new Map();
  List<String> _options = new List();

  final Completer<Answer> _answer = new Completer();

  Future get answer =>  _answer.future;

  Command(String this.command, {Map<String, dynamic> parameters: null, List<String> options: null}){
    if(parameters != null){
      _parameters = parameters;
    }
    if(options != null){
      _options = options;
    }
  }

  void gotAnswer(Answer pA){
    _answer.complete(pA);
  }
  
  void gotAnswerError(Map<String, dynamic> pE){
    _answer.completeError(pE);
  }

  String toString(){
    final StringBuffer lResult = new StringBuffer(command);
    _parameters.forEach((String pKey, dynamic pValue){
      if(pValue is String || pValue is num || pValue is bool){
        lResult
          ..write(" ")
          ..write(Packet.escape(pKey))
          ..write("=")
          ..write(Packet.escape(pValue.toString()));
      } else if(pValue is List){
        final String lKey = Packet.escape(pKey);
        lResult.write(pValue.map((dynamic pItem){
          return "${lKey}=${Packet.escape(pItem.toString())}";
        }).join("|"));
      }
    });
    _options.forEach((String pOption){
      lResult
        ..write(" -")
        ..write(Packet.escape(pOption));
    });
    return lResult.toString();
  }


}

class Answer extends Packet {
  final List<Map<String, String>> items;
  final Map<String, String> status;

  Answer(List this.items, Map this.status);
}

class Notification extends Packet {
  final String name;
  final Map<String, String> parameters;
  
  Notification(String this.name, Map this.parameters);
}